print("import: 'gensim'")
import gensim

print("import: 'gensim.corpora'")
import gensim.corpora

print("import: 'gensim.models'")
import gensim.models

print("import: 'gensim.models.wrappers'")
import gensim.models.wrappers

print("import: 'gensim.parsing'")
import gensim.parsing

print("import: 'gensim.scripts'")
import gensim.scripts

print("import: 'gensim.similarities'")
import gensim.similarities

print("import: 'gensim.sklearn_api'")
import gensim.sklearn_api

print("import: 'gensim.summarization'")
import gensim.summarization

print("import: 'gensim.test'")
import gensim.test

print("import: 'gensim.topic_coherence'")
import gensim.topic_coherence

